import re
import sys
import logging
import yaml
import pytest
from jouvence.document import (
    JouvenceSceneElement, JouvenceSceneSection,
    TYPE_ACTION, TYPE_CENTEREDACTION, TYPE_CHARACTER, TYPE_DIALOG,
    TYPE_PARENTHETICAL, TYPE_TRANSITION, TYPE_LYRICS, TYPE_PAGEBREAK,
    TYPE_SYNOPSIS,
    _scene_element_type_str)
from jouvence.parser import JouvenceParser, JouvenceParserError


def pytest_addoption(parser):
    parser.addoption(
            '--log-debug',
            action='store_true',
            help="Sets the Jouvence logger to output debug info to stdout.")


def pytest_configure(config):
    if config.getoption('--log-debug'):
        hdl = logging.StreamHandler(stream=sys.stdout)
        logging.getLogger('jouvence').addHandler(hdl)
        logging.getLogger('jouvence').setLevel(logging.DEBUG)


def pytest_collect_file(parent, path):
    if path.ext == '.yaml' and path.basename.startswith("test"):
        return JouvenceScriptTestFile(path, parent)
    return None


def assert_scenes(actual, scenes):
    assert len(actual) == len(scenes)
    for a, e in zip(actual, scenes):
        assert_scene(a, e[0], e[1:])


def assert_scene(actual, header, paragraphs):
    if header is not None:
        assert actual.header == header
    assert len(actual.paragraphs) == len(paragraphs)
    for a, e in zip(actual.paragraphs, paragraphs):
        assert_paragraph(a, e)


def assert_paragraph(actual, expected):
    if isinstance(expected, str):
        assert isinstance(actual, JouvenceSceneElement)
        assert actual.type == TYPE_ACTION
        assert actual.text == expected
    elif isinstance(expected, JouvenceSceneElement):
        assert isinstance(actual, JouvenceSceneElement)
        assert actual.type == expected.type
        assert actual.text == expected.text
    else:
        raise NotImplementedError("Don't know what this is: %s" % expected)


def _c(name):
    return JouvenceSceneElement(TYPE_CHARACTER, name)


def _p(text):
    return JouvenceSceneElement(TYPE_PARENTHETICAL, text)


def _d(text):
    return JouvenceSceneElement(TYPE_DIALOG, text)


def _t(text):
    return JouvenceSceneElement(TYPE_TRANSITION, text)


def _l(text):
    return JouvenceSceneElement(TYPE_LYRICS, text)


class UnexpectedScriptOutput(Exception):
    def __init__(self, actual, expected):
        self.actual = actual
        self.expected = expected


class JouvenceScriptTestFile(pytest.File):
    def collect(self):
        spec = yaml.load_all(self.fspath.open(encoding='utf8'))
        for i, item in enumerate(spec):
            name = '%s_%d' % (self.fspath.basename, i)
            if 'test_name' in item:
                name += '_%s' % item['test_name']
            yield JouvenceScriptTestItem(name, self, item)


class JouvenceScriptTestItem(pytest.Item):
    def __init__(self, name, parent, spec):
        super().__init__(name, parent)
        self.spec = spec

    def reportinfo(self):
        return self.fspath, 0, "jouvence script test: %s" % self.name

    def runtest(self):
        intext = self.spec.get('in')
        expected = self.spec.get('out')
        title = self.spec.get('title')
        if intext is None or expected is None:
            raise Exception("No 'in' or 'out' specified.")

        parser = JouvenceParser()
        doc = parser.parseString(intext)
        if title is not None:
            assert title == doc.title_values

        exp_scenes = make_scenes(expected)
        try:
            assert_scenes(doc.scenes, exp_scenes)
        except AssertionError:
            raise UnexpectedScriptOutput(doc.scenes, exp_scenes)

    def repr_failure(self, excinfo):
        if isinstance(excinfo.value, JouvenceParserError):
            return ('\n'.join(
                ['Parser error:', str(excinfo.value)]))
        if isinstance(excinfo.value, UnexpectedScriptOutput):
            return ('\n'.join(
                ['Unexpected output:'] +
                ['', 'Actual:'] +
                list(_repr_doc_scenes(excinfo.value.actual)) +
                ['', 'Expected:'] +
                list(_repr_expected_scenes(excinfo.value.expected))))
        return super().repr_failure(excinfo)


def _repr_doc_scenes(scenes):
    for s in scenes:
        yield 'Scene: "%s"' % s.header
        for p in s.paragraphs:
            yield '  %s: "%s"' % (_scene_element_type_str(p.type),
                                  p.text)


def _repr_expected_scenes(scenes):
    for s in scenes:
        yield 'Scene: "%s"' % s[0]
        for p in s[1:]:
            if isinstance(p, str):
                yield '  ACTION: "%s"' % p
            else:
                yield '  %s: "%s"' % (_scene_element_type_str(p.type),
                                      p.text)


RE_BLANK_LINE = re.compile(r"^\s+$")


def make_scenes(spec):
    if not isinstance(spec, list):
        raise Exception("Script specs must be lists.")

    out = []
    cur_header = None
    cur_paras = []

    for item in spec:
        if item == '<pagebreak>':
            cur_paras.append(JouvenceSceneElement(TYPE_PAGEBREAK, None))
            continue

        token = item[:1]
        if token == '.':
            if cur_header or cur_paras:
                out.append([cur_header] + cur_paras)
            cur_header = item[1:]
            cur_paras = []
        elif token == '!':
            if item[1:3] == '><':
                cur_paras.append(
                    JouvenceSceneElement(TYPE_CENTEREDACTION, item[3:]))
            else:
                cur_paras.append(item[1:])
        elif token == '@':
            cur_paras.append(_c(item[1:]))
        elif token == '=':
            cur_paras.append(_d(item[1:]))
        elif token == '_':
            cur_paras.append(_p(item[1:]))
        elif token == '>':
            cur_paras.append(_t(item[1:]))
        elif token == '~':
            cur_paras.append(_l(item[1:]))
        elif token == '#':
            cur_paras.append(
                JouvenceSceneSection(1, item[1:]))
        elif token == '+':
            cur_paras.append(
                JouvenceSceneElement(TYPE_SYNOPSIS, item[1:]))
        else:
            raise Exception("Unknown token: %s" % token)
    if cur_header or cur_paras:
        out.append([cur_header] + cur_paras)
    return out



import sys


class JouvenceDocument:
    """Represents a Fountain screenplay in a structured way.

    A screenplay contains:

    * A title page (optional) with a key/value dictionary of settings.
    * A list of scenes.
    * Each scene contains a list of paragraphs of various types.
    """
    def __init__(self):
        self.title_values = {}
        self.scenes = []

    def addScene(self, header=None):
        """Adds a scene with the specified header."""
        s = JouvenceScene()
        if header:
            s.header = header
        self.scenes.append(s)
        return s

    def lastScene(self, auto_create=True):
        """Gets the last scene in the screenplay.

            `auto_create`
                If ``True``, and the screenplay has no scenes, create
                a scene with an empty header text. Otherwise, return
                ``None``.
        """
        try:
            return self.scenes[-1]
        except IndexError:
            if auto_create:
                s = self.addScene()
                return s
            return None

    def lastParagraph(self):
        """Gets the last paragraph of the last scene in the screenplay.

        If there's no scene in the screenplay, return ``None``.
        """
        s = self.lastScene(False)
        if s:
            return s.lastParagraph()
        return None


class JouvenceScene:
    """A scene in a screenplay."""
    def __init__(self):
        self.header = None
        self.paragraphs = []
        self._adders = {}

    def __getattr__(self, name):
        if name.startswith('add'):
            add_type_name = name[3:]
            try:
                adder = self._adders[add_type_name]
            except KeyError:
                module = sys.modules[__name__]
                add_type = getattr(module,
                                   'TYPE_%s' % add_type_name.upper())

                def _type_adder(_text):
                    new_p = JouvenceSceneElement(add_type, _text)
                    self.paragraphs.append(new_p)
                    return new_p

                adder = _type_adder
                self._adders[add_type_name] = adder
            return adder
        else:
            raise AttributeError

    def addPageBreak(self):
        """Adds a page break (paragraph with ``TYPE_PAGEBREAK`` type)."""
        self.paragraphs.append(JouvenceSceneElement(TYPE_PAGEBREAK, None))

    def addSection(self, depth, text):
        """Adds a section (a :class:`~JouvenceSceneSection` instance)."""
        self.paragraphs.append(JouvenceSceneSection(depth, text))

    def lastParagraph(self):
        try:
            return self.paragraphs[-1]
        except IndexError:
            return None


class JouvenceSceneElement:
    """An element of a screenplay scene, _e.g._ an action, a dialogue
    line, a parenthetical, etc.
    """
    def __init__(self, el_type, text):
        self.type = el_type
        self.text = text

    def __str__(self):
        return '%s: %s' % (
            _scene_element_type_str(self.type),
            _ellipsis(self.text, 15))


class JouvenceSceneSection(JouvenceSceneElement):
    def __init__(self, depth, text):
        super().__init__(TYPE_SECTION, text)
        self.depth = depth


TYPE_ACTION = 0
TYPE_CENTEREDACTION = 1
TYPE_CHARACTER = 2
TYPE_DIALOG = 3
TYPE_PARENTHETICAL = 4
TYPE_TRANSITION = 5
TYPE_LYRICS = 6
TYPE_PAGEBREAK = 7
TYPE_SECTION = 8
TYPE_SYNOPSIS = 9


def _scene_element_type_str(t):
    if t == TYPE_ACTION:
        return 'ACTION'
    if t == TYPE_CENTEREDACTION:
        return 'CENTEREDACTION'
    if t == TYPE_CHARACTER:
        return 'CHARACTER'
    if t == TYPE_DIALOG:
        return 'DIALOG'
    if t == TYPE_PARENTHETICAL:
        return 'PARENTHETICAL'
    if t == TYPE_TRANSITION:
        return 'TRANSITION'
    if t == TYPE_LYRICS:
        return 'LYRICS'
    if t == TYPE_PAGEBREAK:
        return 'PAGEBREAK'
    if t == TYPE_SECTION:
        return 'SECTION'
    if t == TYPE_SYNOPSIS:
        return 'SYNOPSIS'
    raise NotImplementedError()


def _ellipsis(text, length):
    if len(text) > length:
        return text[:length - 3] + '...'
    return text



"""The ``jouvence.html`` module contains utilities for turning a
Fountain screenplay into a web page, along with other web-related
features.
"""
import os.path
from markupsafe import escape
from .renderer import BaseDocumentRenderer, BaseTextRenderer


def _elem(out, elem_name, class_name, contents):
    f = out.write
    f('<%s' % elem_name)
    if class_name:
        f(' class="jouvence-%s"' % class_name)
    f('>')
    f(contents)
    f('</%s>\n' % elem_name)


def _br(text, strip_first=False):
    lines = text.split('\n')
    if strip_first and lines[0].strip() == '':
        lines = lines[1:]
    return '<br/>\n'.join(lines)


def _res(filename):
    path = os.path.join(os.path.dirname(__file__), 'resources', filename)
    with open(path, 'r') as fp:
        return fp.read()


def get_css():
    """Gets the default CSS screenplay theme that Jouvence uses."""
    return _res('html_styles.css')


class HtmlDocumentRenderer(BaseDocumentRenderer):
    """A document renderer that formats a screenplay into a web page.
    """
    def __init__(self, standalone=True):
        """Creates a new HTML renderer.

            `standalone`
                If ``True`` (the default), it will render a complete HTML
                page, with CSS styling. If ``False``, it will just produce
                the markup for the screenplay itself. The calling
                application is responsible for producing the rest of the
                HTML page, along with CSS code, to render the screenplay
                properly.
        """
        super().__init__(HtmlTextRenderer())
        self.standalone = standalone

    def get_css(self):
        """Gets the default CSS screenplay theme that Jouvence uses.

        This is the same as :func:`~jouvence.html.get_css`.
        """
        return _res('html_styles.css')

    def write_header(self, doc, out):
        if self.standalone:
            meta = doc.title_values.get
            data = {
                # TODO: need a "strip formatting" to have a clean title.
                'title': meta('title', "Fountain Screenplay"),
                'description': meta('description', ''),
                'css': self.get_css()
            }
            out.write(_res('html_header.html') % data)
        out.write('<div class="jouvence-doc">\n')
        out.write('<div class="jouvence-main">\n')

    def write_footer(self, doc, out):
        out.write('</div>\n')  # End `jouvence-main`

        if self.text_renderer.notes:
            out.write('<div class="jouvence-notes">\n')
            self._render_footnotes(out)
            out.write('</div>\n')  # End `jouvence-notes`

        out.write('</div>\n')  # End `jouvence-doc`

        if self.standalone:
            out.write(_res('html_footer.html'))

    def write_title_page(self, values, out):
        out.write('<div class="jouvence-title-page">\n')

        _elem(out, 'h1', None, _br(values['title']))
        _elem(out, 'p', 'title-page-heading', _br(values['credit']))
        _elem(out, 'p', 'title-page-heading', _br(values['author']))

        ddate = values.get('date') or values.get('draft date')
        if ddate:
            _elem(out, 'p', 'title-page-footer', _br(ddate))
        contact = values.get('contact')
        if contact:
            _elem(out, 'p', 'title-page-footer', _br(contact))

        out.write('</div>\n')
        self.write_pagebreak(out)

    def write_scene_heading(self, text, out):
        _elem(out, 'p', 'scene-heading', text)

    def write_action(self, text, out):
        _elem(out, 'p', 'action', _br(text, True))

    def write_centeredaction(self, text, out):
        _elem(out, 'p', 'action-centered', _br(text, True))

    def write_character(self, text, out):
        _elem(out, 'p', 'character', text)

    def write_dialog(self, text, out):
        _elem(out, 'p', 'dialog', _br(text))

    def write_parenthetical(self, text, out):
        _elem(out, 'p', 'parenthetical', text)

    def write_transition(self, text, out):
        _elem(out, 'p', 'transition', text)

    def write_lyrics(self, text, out):
        _elem(out, 'p', 'lyrics', _br(text, True))

    def write_pagebreak(self, out):
        out.write('<hr/>\n')

    def write_section(self, depth, text, out):
        _elem(out, 'p', 'section', '%s %s' % ('#' * depth, text))

    def write_synopsis(self, text, out):
        _elem(out, 'p', 'synopsis', text)

    def _render_footnotes(self, out):
        for i, n in enumerate(self.text_renderer.notes):
            note_id = i + 1
            out.write(
                '<div class="jouvence-note" id="jouvence-note-%d">' %
                note_id)
            text = '<sup>%d</sup> %s' % (note_id, n)
            _elem(out, 'p', None, _br(text))
            out.write('</div>\n')


class HtmlTextRenderer(BaseTextRenderer):
    """A text renderer for producing HTML markup."""
    def __init__(self):
        self.notes = []

    def render_text(self, text):
        return super().render_text(escape(text))

    def make_italics(self, text):
        return '<em>%s</em>' % text

    def make_bold(self, text):
        return '<strong>%s</strong>' % text

    def make_underline(self, text):
        return '<u>%s</u>' % text

    def make_note(self, text):
        note_id = len(self.notes) + 1
        out = '<sup id="jouvence-note-ref-%d">' % note_id
        out += '<a rel="footnote" href="#jouvence-note-%d">' % note_id
        out += str(note_id)
        out += '</a></sup>'
        self.notes.append(text)
        return out



import re
import logging
from .document import TYPE_ACTION


logger = logging.getLogger(__name__)


class JouvenceState:
    def __init__(self):
        pass

    def match(self, fp, ctx):
        return False

    def consume(self, fp, ctx):
        raise NotImplementedError()

    def exit(self, ctx, next_state):
        pass


class JouvenceParserError(Exception):
    def __init__(self, line_no, message):
        super().__init__("Error line %d: %s" % (line_no, message))


ANY_STATE = object()
EOF_STATE = object()


# Note how boneyard start/end patterns (/* */) by themselves are
# considered an empty line.
RE_EMPTY_LINE = re.compile(r"^(/\*|\*/)?$", re.M)

RE_TITLE_KEY_VALUE = re.compile(r"^(?P<key>[\w\s\-]+)\s*:\s*")


class _TitlePageState(JouvenceState):
    def __init__(self):
        super().__init__()
        self._cur_key = None
        self._cur_val = None

    def consume(self, fp, ctx):
        line = fp.peekline()
        is_match = RE_TITLE_KEY_VALUE.match(line)
        is_line1_empty = RE_EMPTY_LINE.match(line)
        if not is_match:
            logger.debug("No title page value found on line 1.")
            if not is_line1_empty:
                # Add a fake empty line at the beginning of the text if
                # there's not one already. This makes state matching easier.
                fp._addBlankAt0()
            return ANY_STATE

        while True:
            line = fp.readline()
            if not line:
                return EOF_STATE

            m = RE_TITLE_KEY_VALUE.match(line)
            if m:
                # Commit current value, start new one.
                self._commit(ctx)
                self._cur_key = m.group('key').lower()
                self._cur_val = line[m.end():]
            else:
                # Keep accumulating the value of one of the title page's
                # values.
                self._cur_val += line.lstrip()

            if RE_EMPTY_LINE.match(fp.peekline()):
                if (self._cur_key and
                        not self._cur_val and
                        not ctx.document.title_values):
                    # We thought there was a title page, but it turns out
                    # we probably mistakenly matched something like a
                    # transition.
                    logger.debug("Aborting title page parsing, resetting "
                                 "back to first line.")
                    fp.reset()
                    if not is_line1_empty:
                        fp._addBlankAt0()
                    break

                # Finished with the page title, now move on to the first scene.
                self._commit(ctx)
                break

        return ANY_STATE

    def exit(self, ctx, next_state):
        self._commit(ctx)

    def _commit(self, ctx):
        if self._cur_key is not None:
            val = self._cur_val.rstrip('\r\n')
            ctx.document.title_values[self._cur_key] = val
            self._cur_key = None
            self._cur_val = None


RE_SCENE_HEADER_PATTERN = re.compile(
    r"^(int|ext|est|int/ext|int./ext|i/e)[\s\.]", re.I)


class _SceneHeaderState(JouvenceState):
    def match(self, fp, ctx):
        lines = fp.peeklines(3)
        return (
            RE_EMPTY_LINE.match(lines[0]) and
            RE_SCENE_HEADER_PATTERN.match(lines[1]) and
            RE_EMPTY_LINE.match(lines[2]))

    def consume(self, fp, ctx):
        fp.readline()  # Get past the blank line.
        line = fp.readline().rstrip('\r\n')
        line = line.lstrip('.')  # In case it was forced.
        ctx.document.addScene(line)
        return ANY_STATE


class _ActionState(JouvenceState):
    def __init__(self):
        super().__init__()
        self.text = ''

    def match(self, fp, ctx):
        return True

    def consume(self, fp, ctx):
        is_first_line = True
        while True:
            line = fp.readline()
            if not line:
                return EOF_STATE

            if is_first_line:
                # Ignore the fake blank line at 0 if it's threre.
                if fp.line_no == 0:
                    continue

                line = line.lstrip('!')  # In case it was forced.
                is_first_line = False

            # If the next line is empty, strip the carriage return from
            # the line we just got because it's probably gonna be the
            # last one.
            if RE_EMPTY_LINE.match(fp.peekline()):
                self.text += line.rstrip("\r\n")
                break
            # ...otherwise, add the line with in full.
            self.text += line

        return ANY_STATE

    def exit(self, ctx, next_state):
        last_para = ctx.document.lastParagraph()
        if last_para and last_para.type == TYPE_ACTION:
            last_para.text += '\n' + self.text
        else:
            ctx.document.lastScene().addAction(self.text)


RE_CENTERED_LINE = re.compile(r"^\s*>\s*.*\s*<\s*$")


class _CenteredActionState(JouvenceState):
    def __init__(self):
        super().__init__()
        self.text = ''
        self._aborted = False

    def match(self, fp, ctx):
        lines = fp.peeklines(2)
        return (
            RE_EMPTY_LINE.match(lines[0]) and
            RE_CENTERED_LINE.match(lines[1]))

    def consume(self, fp, ctx):
        snapshot = fp.snapshot()
        fp.readline()  # Get past the empty line.
        while True:
            line = fp.readline()
            if not line:
                return EOF_STATE

            clean_line = line.rstrip('\r\n')
            eol = line[len(clean_line):]

            clean_line = clean_line.strip()
            if clean_line[0] != '>' or clean_line[-1] != '<':
                # The whole paragraph must have `>` and `<` wrappers, so
                # if we detect a line that doesn't have them, we make this
                # paragraph be a normal action instead.
                fp.restore(snapshot)
                self._aborted = True
                return _ActionState()
            else:
                # Remove wrapping `>`/`<`, and spaces.
                clean_line = clean_line[1:-1].strip()

            if RE_EMPTY_LINE.match(fp.peekline()):
                self.text += clean_line
                break
            self.text += clean_line + eol

        return ANY_STATE

    def exit(self, ctx, next_state):
        if not self._aborted:
            ctx.document.lastScene().addCenteredAction(self.text)


RE_CHARACTER_LINE = re.compile(r"^\s*[A-Z][A-Z\-\._\s]+\s*(\(.*\))?$")


class _CharacterState(JouvenceState):
    def match(self, fp, ctx):
        lines = fp.peeklines(3)
        return (RE_EMPTY_LINE.match(lines[0]) and
                RE_CHARACTER_LINE.match(lines[1]) and
                not RE_EMPTY_LINE.match(lines[2]))

    def consume(self, fp, ctx):
        fp.readline()  # Get past the empty line.
        line = fp.readline().rstrip('\r\n')
        line = line.lstrip()  # Remove indenting.
        line = line.lstrip('@')  # In case it was forced.
        ctx.document.lastScene().addCharacter(line)
        return [_ParentheticalState, _DialogState]


RE_PARENTHETICAL_LINE = re.compile(r"^\s*\(.*\)\s*$")


class _ParentheticalState(JouvenceState):
    def match(self, fp, ctx):
        # We only get here from a `_CharacterState` so we know the previous
        # one is already that.
        line = fp.peekline()
        return RE_PARENTHETICAL_LINE.match(line)

    def consume(self, fp, ctx):
        line = fp.readline().lstrip().rstrip('\r\n')
        ctx.document.lastScene().addParenthetical(line)

        next_line = fp.peekline()
        if not RE_EMPTY_LINE.match(next_line):
            return _DialogState()

        return ANY_STATE


class _DialogState(JouvenceState):
    def __init__(self):
        super().__init__()
        self.text = ''

    def match(self, fp, ctx):
        # We only get here from a `_CharacterState` or `_ParentheticalState`
        # so we just need to check there's some text.
        line = fp.peekline()
        return not RE_EMPTY_LINE.match(line)

    def consume(self, fp, ctx):
        while True:
            line = fp.readline()
            if not line:
                return EOF_STATE

            line = line.lstrip()  # Remove indenting.

            # Next we could be either continuing the dialog line, going to
            # a parenthetical, or exiting dialog altogether.
            next_line = fp.peekline()

            if RE_PARENTHETICAL_LINE.match(next_line):
                self.text += line.rstrip('\r\n')
                return _ParentheticalState()

            if RE_EMPTY_LINE.match(next_line):
                self.text += line.rstrip('\r\n')
                break
            self.text += line

        return ANY_STATE

    def exit(self, ctx, next_state):
        ctx.document.lastScene().addDialog(self.text.rstrip('\r\n'))


class _LyricsState(JouvenceState):
    def __init__(self):
        super().__init__()
        self.text = ''
        self._aborted = False

    # No `match` method, this can only be forced.
    # (see `_ForcedParagraphStates`)

    def consume(self, fp, ctx):
        snapshot = fp.snapshot()
        fp.readline()  # Get past the empty line.
        while True:
            line = fp.readline()
            if not line:
                return EOF_STATE

            if line.startswith('~'):
                line = line.lstrip('~')
            else:
                logger.debug("Rolling back lyrics into action paragraph.")
                fp.restore(snapshot)
                self._aborted = True
                return _ActionState()

            if RE_EMPTY_LINE.match(fp.peekline()):
                self.text += line.rstrip('\r\n')
                break
            self.text += line

        return ANY_STATE

    def exit(self, ctx, next_state):
        if not self._aborted:
            ctx.document.lastScene().addLyrics(self.text)


RE_TRANSITION_LINE = re.compile(r"^\s*[^a-z]+TO\:$")


class _TransitionState(JouvenceState):
    def match(self, fp, ctx):
        lines = fp.peeklines(3)
        return (
            RE_EMPTY_LINE.match(lines[0]) and
            RE_TRANSITION_LINE.match(lines[1]) and
            RE_EMPTY_LINE.match(lines[2]))

    def consume(self, fp, ctx):
        fp.readline()  # Get past the empty line.
        line = fp.readline().lstrip().rstrip('\r\n')
        line = line.lstrip('>')  # In case it was forced.
        ctx.document.lastScene().addTransition(line)
        return ANY_STATE


RE_PAGE_BREAK_LINE = re.compile(r"^\=\=\=+$")


class _PageBreakState(JouvenceState):
    def match(self, fp, ctx):
        lines = fp.peeklines(3)
        return (
            RE_EMPTY_LINE.match(lines[0]) and
            RE_PAGE_BREAK_LINE.match(lines[1]) and
            RE_EMPTY_LINE.match(lines[2]))

    def consume(self, fp, ctx):
        fp.readline()
        fp.readline()
        ctx.document.lastScene().addPageBreak()
        return ANY_STATE


class _ForcedParagraphStates(JouvenceState):
    STATE_SYMBOLS = {
        '.': _SceneHeaderState,
        '!': _ActionState,
        '@': _CharacterState,
        '~': _LyricsState,
        '>': _TransitionState
    }

    def __init__(self):
        super().__init__()
        self._state_cls = None
        self._consume_empty_line = False

    def match(self, fp, ctx):
        lines = fp.peeklines(2)
        symbol = lines[1][:1]
        if (RE_EMPTY_LINE.match(lines[0]) and
                symbol in self.STATE_SYMBOLS):
            # Special case: don't force a transition state if it's
            # really some centered text.
            if symbol == '>' and RE_CENTERED_LINE.match(lines[1]):
                return False

            self._state_cls = self.STATE_SYMBOLS[symbol]

            # Special case: for forced action paragraphs, don't leave
            # the blank line there.
            if symbol == '!':
                self._consume_empty_line = True

            return True
        return False

    def consume(self, fp, ctx):
        if self._consume_empty_line:
            fp.readline()
        return self._state_cls()


RE_BONEYARD_START = re.compile(r"^/\*")
RE_BONEYARD_END = re.compile(r"\*/\s*$")


class _BoneyardState(JouvenceState):
    def match(self, fp, ctx):
        return RE_BONEYARD_START.match(fp.peekline())

    def consume(self, fp, ctx):
        while True:
            fp.readline()
            if RE_BONEYARD_END.match(fp.peekline()):
                break
        return ANY_STATE


RE_SECTION_LINE = re.compile(r"^(?P<depth>#+)\s*")


class _SectionState(JouvenceState):
    def match(self, fp, ctx):
        lines = fp.peeklines(3)
        return (RE_EMPTY_LINE.match(lines[0]) and
                RE_SECTION_LINE.match(lines[1]) and
                RE_EMPTY_LINE.match(lines[2]))

    def consume(self, fp, ctx):
        fp.readline()
        line = fp.readline()
        m = RE_SECTION_LINE.match(line)
        depth = len(m.group('depth'))
        line = line[m.end():].rstrip('\r\n')
        ctx.document.lastScene().addSection(depth, line)
        return ANY_STATE


RE_SYNOPSIS_LINE = re.compile(r"^=[^=]")


class _SynopsisState(JouvenceState):
    def match(self, fp, ctx):
        lines = fp.peeklines(3)
        return (RE_EMPTY_LINE.match(lines[0]) and
                RE_SYNOPSIS_LINE.match(lines[1]) and
                RE_EMPTY_LINE.match(lines[2]))

    def consume(self, fp, ctx):
        fp.readline()
        line = fp.readline()
        line = line[1:].lstrip()  # Remove the `#`, and the following spaces.
        line = line.rstrip('\r\n')
        ctx.document.lastScene().addSynopsis(line)
        return ANY_STATE


class _EmptyLineState(JouvenceState):
    def __init__(self):
        super().__init__()
        self.line_count = 0

    def match(self, fp, ctx):
        return RE_EMPTY_LINE.match(fp.peekline())

    def consume(self, fp, ctx):
        line = fp.readline()
        # Increment the number of empty lines to add to the current action,
        # but:
        # - don't take into account the fake blank at 0
        # - don't take into account boneyard endings
        if (fp.line_no > 1 and
                not RE_BONEYARD_END.match(line)):
            self.line_count += 1
        return ANY_STATE

    def exit(self, ctx, next_state):
        if self.line_count > 0:
            text = self.line_count * '\n'
            last_para = ctx.document.lastParagraph()
            if last_para and last_para.type == TYPE_ACTION:
                last_para.text += text
            else:
                ctx.document.lastScene().addAction(text[1:])


ROOT_STATES = [
    _ForcedParagraphStates,  # Must be first.
    _SceneHeaderState,
    _CharacterState,
    _TransitionState,
    _PageBreakState,
    _CenteredActionState,
    _SectionState,
    _SynopsisState,
    _BoneyardState,
    _EmptyLineState,   # Must be second to last.
    _ActionState,  # Must be last.
]


class _PeekableFile:
    def __init__(self, fp):
        self.line_no = 1
        self._fp = fp
        # This is for adding a "fake" blank line at the beginning of the
        # file, to help with match things on the first line.
        # (has blank line, is blank line unread)
        self._blankAt0 = (False, False)

    def readline(self):
        if self._blankAt0[1]:
            self._blankAt0 = (True, False)
            self.line_no = 0
            return '\n'

        data = self._fp.readline()
        self.line_no += 1
        return data

    def peekline(self):
        if self._blankAt0[1]:
            return '\n'

        pos = self._fp.tell()
        line = self._fp.readline()
        self._fp.seek(pos)
        return line

    def peeklines(self, count):
        pos = self._fp.tell()
        lines = []
        if self._blankAt0[1]:
            lines.append('\n')
            count -= 1
        for i in range(count):
            lines.append(self._fp.readline())
        self._fp.seek(pos)
        return lines

    def snapshot(self):
        return (self._fp.tell(), self._blankAt0, self.line_no)

    def restore(self, snapshot):
        self._fp.seek(snapshot[0])
        self._blankAt0 = snapshot[1]
        self.line_no = snapshot[2]

    def reset(self):
        self._fp.seek(0)
        if self._blankAt0[0]:
            self._blankAt0 = (True, True)
            self.line_no = 0
        else:
            self._blankAt0 = (False, False)
            self.line_no = 1

    def _addBlankAt0(self):
        if self._fp.tell() != 0:
            raise Exception(
                "Can't add blank line at 0 if reading has started.")
        self._blankAt0 = (True, True)
        self.line_no = 0


class _JouvenceStateMachine:
    def __init__(self, fp, doc):
        self.fp = _PeekableFile(fp)
        self.state = None
        self.document = doc

    @property
    def line_no(self):
        return self.fp.line_no

    def run(self):
        # Start parsing! Here we try to do a mostly-forward-only parser with
        # non overlapping regexes to make it decently fast.
        self.state = _TitlePageState()
        while True:
            logger.debug("State '%s' consuming from '%s'..." %
                         (self.state.__class__.__name__, self.fp.peekline()))
            res = self.state.consume(self.fp, self)

            # See if we reached the end of the file.
            if not self.fp.peekline():
                logger.debug("Reached end of line... ending parsing.")
                res = EOF_STATE

            # Figure out what to do next...

            if res is None:
                raise JouvenceParserError(
                    self.line_no,
                    "State '%s' returned a `None` result. "
                    "States need to return `ANY_STATE`, one or more specific "
                    "states, or `EOF_STATE` if they reached the end of the "
                    "file." % self.state.__class__.__name__)

            elif res is ANY_STATE or isinstance(res, list):
                # State wants to exit, we need to figure out what is the
                # next state.
                pos = self.fp._fp.tell()
                next_states = res
                if next_states is ANY_STATE:
                    next_states = ROOT_STATES
                logger.debug("Trying to match next state from: %s" %
                             [t.__name__ for t in next_states])
                for sc in next_states:
                    s = sc()
                    if s.match(self.fp, self):
                        logger.debug("Matched state %s" %
                                     s.__class__.__name__)
                        self.fp._fp.seek(pos)
                        res = s
                        break
                else:
                    raise Exception("Can't match following state after: %s" %
                                    self.state)

                # Handle the current state before we move on to the new one.
                if self.state:
                    self.state.exit(self, res)
                self.state = res

            elif isinstance(res, JouvenceState):
                # State wants to exit, wants a specific state to be next.
                if self.state:
                    self.state.exit(self, res)
                self.state = res

            elif res is EOF_STATE:
                # Reached end of file.
                if self.state:
                    self.state.exit(self, res)
                break

            else:
                raise Exception("Unsupported state result: %s" % res)


class JouvenceParser:
    """ Parses a Fountain document and returns a
    :class:`~jouvence.document.JouvenceDocument` instance.
    """
    def __init__(self):
        pass

    def parse(self, filein):
        """Parses a file or stream. This must either be a Python file object,
        or the path to file on disk. Returns a
        :class:`~jouvence.document.JouvenceDocument` instance.
        """
        if isinstance(filein, str):
            with open(filein, 'r') as fp:
                return self._doParse(fp)
        else:
            return self._doParse(fp)

    def parseString(self, text):
        """Parses a string. Returns a
        :class:`~jouvence.document.JouvenceDocument` instance.
        """
        import io
        with io.StringIO(text) as fp:
            return self._doParse(fp)

    def _doParse(self, fp):
        from .document import JouvenceDocument
        doc = JouvenceDocument()
        machine = _JouvenceStateMachine(fp, doc)
        machine.run()
        return doc



# -*- coding: utf-8 -*-
"""
    JOUVENCE
    ========

    `Fountain`_ is a plain text markup language for screenwriting. Jouvence
    is a Python package for parsing and rendering Fountain documents.

    Jouvence supports:

    * Most of the Fountain specification (see limitations below).
    * Rendering to HTML and terminals.

    .. _fountain: http://fountain.io/

    :copyright: (c) 2017, BOLT80.
    :license: Apache2. See LICENSE file for details.
"""

__docformat__ = 'restructuredtext en'

try:
    from .version import version
except ImportError:
    version = '<unknown>'

__version__ = version



import re
from jouvence.document import (
    TYPE_ACTION, TYPE_CENTEREDACTION, TYPE_CHARACTER, TYPE_DIALOG,
    TYPE_PARENTHETICAL, TYPE_TRANSITION, TYPE_LYRICS, TYPE_PAGEBREAK,
    TYPE_SECTION, TYPE_SYNOPSIS)


class BaseDocumentRenderer:
    """The base class for document renderers.

    Document renderers are given a document to transform into whatever
    format they're responsible for.
    """
    def __init__(self, text_renderer=None):
        """Initializes the document renderer.

            `text_renderer`
                The :class:`~BaseTextRenderer` instance to use for
                formatting text.
        """
        self.force_title_page = False
        self.text_renderer = text_renderer
        if not text_renderer:
            self.text_renderer = NullTextRenderer()

        self._para_rdrs = {
            TYPE_ACTION: self.write_action,
            TYPE_CENTEREDACTION: self.write_centeredaction,
            TYPE_CHARACTER: self.write_character,
            TYPE_DIALOG: self.write_dialog,
            TYPE_PARENTHETICAL: self.write_parenthetical,
            TYPE_TRANSITION: self.write_transition,
            TYPE_LYRICS: self.write_lyrics,
            TYPE_PAGEBREAK: self.write_pagebreak,
            TYPE_SECTION: self.write_section,
            TYPE_SYNOPSIS: self.write_synopsis,
        }

    def _tr(self, text):
        return self.text_renderer.render_text(text)

    def render_doc(self, doc, out):
        """Renders the given document (``doc``) into the given buffer
        (``out``).

        The output buffer must be a file-like object, like the return
        value of the ``open()`` built-in function, or an instance of
        ``io.StringIO``, among others.
        """
        self.write_header(doc, out)
        self.render_title_page(doc.title_values, out)
        for s in doc.scenes:
            self.render_scene(s, out)
        self.write_footer(doc, out)

    def render_title_page(self, values, out):
        if not self.force_title_page and not values:
            return

        clean_values = values.copy()
        clean_values.setdefault('title', 'Untitled Screenplay')
        clean_values.setdefault('credit', 'Written by')
        clean_values.setdefault('author', 'Unknown')
        for k in clean_values:
            clean_values[k] = self._tr(clean_values[k])
        self.write_title_page(clean_values, out)

    def render_scene(self, scene, out):
        if scene.header is not None:
            self.write_scene_heading(scene.header, out)
        for p in scene.paragraphs:
            rdr_func = self._para_rdrs[p.type]
            if p.type == TYPE_PAGEBREAK:
                rdr_func(out)
            elif p.type == TYPE_SECTION:
                rdr_func(p.depth, self._tr(p.text), out)
            else:
                rdr_func(self._tr(p.text), out)

    def write_header(self, doc, out):
        pass

    def write_footer(self, doc, out):
        pass

    def write_title_page(self, values, out):
        raise NotImplementedError()

    def write_scene_heading(self, text, out):
        raise NotImplementedError()

    def write_action(self, text, out):
        raise NotImplementedError()

    def write_centeredaction(self, text, out):
        raise NotImplementedError()

    def write_character(self, text, out):
        raise NotImplementedError()

    def write_dialog(self, text, out):
        raise NotImplementedError()

    def write_parenthetical(self, text, out):
        raise NotImplementedError()

    def write_transition(self, text, out):
        raise NotImplementedError()

    def write_lyrics(self, text, out):
        raise NotImplementedError()

    def write_pagebreak(self, out):
        raise NotImplementedError()

    def write_section(self, depth, text, out):
        raise NotImplementedError()

    def write_synopsis(self, text, out):
        raise NotImplementedError()


RE_ITALICS = re.compile(
    r"(?P<before>^|\s)(?P<esc>\\)?\*(?P<text>.*[^\s\*])\*(?=[^a-zA-Z0-9\*]|$)")
RE_BOLD = re.compile(
    r"(?P<before>^|\s)(?P<esc>\\)?\*\*(?P<text>.*[^\s])\*\*(?=[^a-zA-Z0-9]|$)")
RE_UNDERLINE = re.compile(
    r"(?P<before>^|\s)(?P<esc>\\)?_(?P<text>.*[^\s])_(?=[^a-zA-Z0-9]|$)")

RE_NOTE = re.compile(r"\[\[(?P<text>.+?)\]\]", re.DOTALL)


class BaseTextRenderer:
    """A class responsible for rendering text, including such things as
    italics, bold, or underline.
    """
    def render_text(self, text):
        """Processes the given text, and returns the formatted data."""
        # Replace bold stuff to catch double asterisks.
        text = RE_BOLD.sub(self._do_make_bold, text)
        text = RE_ITALICS.sub(self._do_make_italics, text)
        text = RE_UNDERLINE.sub(self._do_make_underline, text)
        text = RE_NOTE.sub(self._do_make_note, text)

        return text

    def _do_make_italics(self, m):
        if m.group('esc'):
            return m.group('before') + '*' + m.group('text') + '*'
        return (
            m.group('before') +
            self.make_italics(m.group('text')))

    def _do_make_bold(self, m):
        if m.group('esc'):
            return m.group('before') + '**' + m.group('text') + '**'
        return (
            m.group('before') +
            self.make_bold(m.group('text')))

    def _do_make_underline(self, m):
        if m.group('esc'):
            return m.group('before') + '_' + m.group('text') + '_'
        return (
            m.group('before') +
            self.make_underline(m.group('text')))

    def _do_make_note(self, m):
        return self.make_note(m.group('text'))

    def make_italics(self, text):
        raise NotImplementedError()

    def make_bold(self, text):
        raise NotImplementedError()

    def make_underline(self, text):
        raise NotImplementedError()

    def make_note(self, text):
        raise NotImplementedError()


class NullTextRenderer(BaseTextRenderer):
    def make_bold(self, text):
        return text

    def make_italics(self, text):
        return text

    def make_underline(self, text):
        return text

    def make_note(self, text):
        return text



import pytest
from jouvence.renderer import BaseTextRenderer


class TestTextRenderer(BaseTextRenderer):
    def make_bold(self, text):
        return 'B:' + text + ':B'

    def make_italics(self, text):
        return 'I:' + text + ':I'

    def make_underline(self, text):
        return 'U:' + text + ':U'

    def make_note(self, text):
        return 'N:' + text + ':N'


@pytest.mark.parametrize('intext, expected', [
    ("_Underline_", "U:Underline:U"),
    ("Here's an _underline_.", "Here's an U:underline:U."),
    ("Here's an _underline_ too", "Here's an U:underline:U too"),
    ("This is not_underline_", "This is not_underline_"),
    ("This is not _underline_either", "This is not _underline_either"),
    ("This is _two underlined_ words.", "This is U:two underlined:U words."),
    ("This is _three underlined words_.",
     "This is U:three underlined words:U."),
    ("This is an \_escaped_ one.", "This is an _escaped_ one.")
])
def test_underline(intext, expected):
    r = TestTextRenderer()
    out = r.render_text(intext)
    assert out == expected


@pytest.mark.parametrize('intext, expected', [
    ("*Italics*", "I:Italics:I"),
    ("Here's some *italics*.", "Here's some I:italics:I."),
    ("Here's some *italics* too", "Here's some I:italics:I too"),
    ("This is not*italics*", "This is not*italics*"),
    ("This is not *italics*either", "This is not *italics*either"),
    ("This is *two italics* words.", "This is I:two italics:I words."),
    ("This is *three italics words*.",
     "This is I:three italics words:I."),
    ("This is some \*escaped* one.", "This is some *escaped* one.")
])
def test_italics(intext, expected):
    r = TestTextRenderer()
    out = r.render_text(intext)
    assert out == expected


@pytest.mark.parametrize('intext, expected', [
    ("**Bold**", "B:Bold:B"),
    ("Here's some **bold**.", "Here's some B:bold:B."),
    ("Here's some **bold** too", "Here's some B:bold:B too"),
    ("This is not**bold**", "This is not**bold**"),
    ("This is not **bold**either", "This is not **bold**either"),
    ("This is **two bold** words.", "This is B:two bold:B words."),
    ("This is **three bold words**.",
     "This is B:three bold words:B."),
    ("This is some \**escaped** one.", "This is some **escaped** one.")
])
def test_bold(intext, expected):
    r = TestTextRenderer()
    out = r.render_text(intext)
    assert out == expected


def test_note():
    r = TestTextRenderer()
    out = r.render_text(
        "This is JACK[[Do we have a better name?]]. He likes movies.")
    expected = "This is JACKN:Do we have a better name?:N. He likes movies."
    assert out == expected


def test_note_with_line_break():
    r = TestTextRenderer()
    out = r.render_text(
        "This is JACK[[Do we have a better name?\n"
        "I think we did]]. He likes movies.")
    expected = ("This is JACKN:Do we have a better name?\n"
                "I think we did:N. He likes movies.")
    assert out == expected


def test_note_multiple():
    r = TestTextRenderer()
    out = r.render_text(
        "This is JACK[[Do we have a better name?]]. "
        "He likes movies[[or plays?]].")
    expected = ("This is JACKN:Do we have a better name?:N. "
                "He likes moviesN:or plays?:N.")
    assert out == expected



"""The ``jouvence.console`` module contains utilities for rendering
a Foutain screenplay in a terminal.
"""
import os
import colorama
from .renderer import BaseDocumentRenderer, BaseTextRenderer


def _w(out, style, text, reset_all=False):
    f = out.write
    f(style)
    f(text)
    if not reset_all:
        f(colorama.Style.NORMAL)
    else:
        f(colorama.Style.RESET_ALL)
    f(os.linesep)


class ConsoleDocumentRenderer(BaseDocumentRenderer):
    """A document renderer that prints a screenplay into a terminal with
    some nice syntax highlighting.
    """
    def __init__(self, width=80):
        super().__init__(ConsoleTextRenderer())
        self.width = width
        colorama.init()

    def write_title_page(self, values, out):
        known = ['title', 'credit', 'author', 'authors', 'source']
        center_values = [values.get(i) for i in known]

        print("", file=out)
        for val in center_values:
            if val is not None:
                for l in val.split('\n'):
                    print(l.center(self.width), file=out)
            print("", file=out)
        print("", file=out)
        print("", file=out)

        known = ['date', 'draft date', 'contact', 'copyright']
        bottom_lines = [values.get(i) for i in known]

        _w(out, colorama.Style.DIM, '\n\n'.join([
            b for b in bottom_lines if b is not None]))
        print("", file=out)
        _w(out, colorama.Style.DIM, 80 * '=')

    def write_scene_heading(self, text, out):
        print("", file=out)
        _w(out, colorama.Fore.WHITE + colorama.Style.BRIGHT, text, True)

    def write_action(self, text, out):
        print(text, file=out)

    def write_centeredaction(self, text, out):
        print("", file=out)
        for line in text.split('\n'):
            print(line.center(self.width), file=out)

    def write_character(self, text, out):
        print("", file=out)
        _w(out, colorama.Fore.WHITE, "\t\t\t" + text, True)

    def write_dialog(self, text, out):
        for line in text.split('\n'):
            print("\t" + line, file=out)

    def write_parenthetical(self, text, out):
        for line in text.split('\n'):
            print("\t\t" + line, file=out)

    def write_transition(self, text, out):
        print("", file=out)
        print("\t\t\t\t" + text, file=out)

    def write_lyrics(self, text, out):
        print("", file=out)
        _w(out, colorama.Fore.MAGENTA, text, True)

    def write_pagebreak(self, out):
        print("", file=out)
        _w(out, colorama.Style.DIM, 80 * '=')

    def write_section(self, depth, text, out):
        print("", file=out)
        _w(out, colorama.Fore.CYAN, '#' * depth + ' ' + text, True)

    def write_synopsis(self, text, out):
        print("", file=out)
        _w(out, colorama.Fore.GREEN, '= ' + text, True)


class ConsoleTextRenderer(BaseTextRenderer):
    """A text renderer for producing ANSI colouring."""
    def _writeStyled(self, style, text):
        return style + text + colorama.Style.NORMAL

    def make_italics(self, text):
        return self._writeStyled(colorama.Style.BRIGHT, text)

    def make_bold(self, text):
        return self._writeStyled(colorama.Style.BRIGHT, text)

    def make_underline(self, text):
        return self._writeStyled(colorama.Style.BRIGHT, text)

    def make_note(self, text):
        out = colorama.Style.DIM + colorama.Fore.MAGENTA
        out += ' [[ ' + text + ' ]] '
        out += colorama.Style.RESET_ALL
        return out



import sys
import argparse


def main():
    """The Jouvence CLI utility.

    It lets the user print a screenplay to the terminal (with syntax
    highlighting), or save a screenplay into a nicely formatted HTML page.
    """
    parser = argparse.ArgumentParser(
        description='Jouvence command line utility')
    parser.add_argument('script')
    parser.add_argument('out_file', nargs='?')
    args = parser.parse_args()

    from jouvence.parser import JouvenceParser
    p = JouvenceParser()
    doc = p.parse(args.script)

    if not args.out_file:
        from jouvence.console import ConsoleDocumentRenderer
        rdr = ConsoleDocumentRenderer()
        rdr.render_doc(doc, sys.stdout)
    else:
        from jouvence.html import HtmlDocumentRenderer
        rdr = HtmlDocumentRenderer()
        with open(args.out_file, 'w') as fp:
            rdr.render_doc(doc, fp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path
from setuptools import setup, find_packages


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fp:
        return fp.read()


def readlines(fname):
    return [l.strip() for l in read(fname).strip().splitlines()]


install_requires = readlines('requirements.txt')
tests_require = readlines('dev-requirements.txt')
long_description = read('README.rst')


setup(
    name="Jouvence",
    use_scm_version={'write_to': 'jouvence/version.py'},
    description="A library for parsing and rendering Fountain screenplays.",
    long_description=long_description,
    author="Ludovic Chabant",
    author_email="ludovic@chabant.com",
    license="Apache License 2.0",
    url="https://bolt80.com/jouvence",
    keywords='fountain screenplay screenwriting screenwriter',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    setup_requires=['setuptools_scm', 'pytest-runner'],
    tests_require=tests_require,
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'jouvence = jouvence.cli:main'
        ]}
)
